#ifndef NOTEST

#include "trithorpe.h"
#include "compiler.h"

void eq_vectors(tri a, tri b)
{
    int alen= a.aSize();
    int blen= b.aSize();
    EQ( alen, blen );

    for ( int i=0; i<alen; i++) {
	EQ( a.aGet(i), b.aGet(i) );
    }
}
void eq_vector_start(tri a, tri b)
{
    int alen= a.aSize();
    int blen= b.aSize();
    MUST( alen <= blen );

    for ( int i=0; i<alen; i++) {
	EQ( a.aGet(i), b.aGet(i) );
    }
}

void Test_TestingMacros()
{
    EQ(5, 5);
    NE(5, 12);
    MUST(23);
    CANT(0);

#ifndef NO_BAD_TESTS
    ISBAD(EQ(1, 5));
    ISBAD(NE(7, 7));
    ISBAD(MUST(0));
    ISBAD(CANT(88));
#endif
}

void Test_tri_assign()
{
    tri x, y, z, k;
    z.setInt(888);
    k.setInt(1001);

    x.setNewTab();
    x.asTab()->store(k,z);
    EQ( 888, x.asTab()->recall(k).asInt() );

    MUST( x.isTab() );
    MUST( x.peek() );
    y= x;

    MUST( y.isTab() );
    EQ( y.peek(), x.peek() );
    EQ( 888, x.asTab()->recall(k).asInt() );
    EQ( 888, y.asTab()->recall(k).asInt() );

    MUST( z.isInt() );
    MUST( x.peek() );
}

void Test_tri()
{
    EQ(4, sizeof(tri));

    tri x;
    CANT(x.isInt());
    MUST(x.isSym());
    CANT(x.isUnk());
    CANT(x.isTab());
    EQ(NIL_GUTS, x.peek());
    MUST( x == NIL );

    x.setInt(42);
    MUST(x.isInt());
    CANT(x.isSym());
    CANT(x.isUnk());
    CANT(x.isTab());
    EQ(42, x.peek());
    EQ(42, x.asInt());
    MUST( x == tri(42) );

    x.setSym('A', 'B', 'C', 'D');
    CANT(x.isInt());
    MUST(x.isSym());
    CANT(x.isUnk());
    CANT(x.isTab());
    EQ(0x41424344, x.peek());
    MUST( x == tri('A','B','C','D') );

    x.setSym('A', 'B', 'C');
    MUST( x == tri('A','B','C') );

    x.setSym('A', 'B');
    MUST( x == tri('A','B' ) );

    x.setSym('A');
    MUST( x == tri('A') );

    x.setNewTab();
    CANT(x.isInt());
    CANT(x.isSym());
    CANT(x.isUnk());
    MUST(x.isTab());
    EQ(0, x.asTab()->arr.capacity());
    EQ(0, x.asTab()->dict.capacity());
    EQ(0, x.asTab()->arr.size());
    EQ(0, x.asTab()->dict.size());

#ifndef NO_BAD_TESTS
    ISBAD(x.setSym('!', 'B', 'C', 'D').peek());
#endif
}

void Test_tri_array()
{
    tri *v = new tri[99];
    for (int i = 0; i < 99; i++)
	v[i].setNewTab();
    delete[]v;

    tri *w = new tri[99];
    for (int i = 0; i < 99; i++)
	w[i].setNewTab();
    for (int i = 0; i < 99; i++)
	w[i].setInt(0);
    delete[]w;
}

void Test_vec()
{
    vec a;
    for (int i = 0; i < 66; i++)
	a[i].setInt(i);
    tri x;
    x.setNewTab();
    for (int i = 0; i < 66; i++)
	a[i].setTab(x.asTab());
}



void Test_tab()
{
    tri x;
    x.setNewTab();
    tab *tp = x.asTab();
    for (int i = 0; i < 111; i++)
	tp->arr[i].setInt(i * i);
    const tab *ctp = x.asTab();
    for (int i = 0; i < 111; i++)
	EQ(i * i, ctp->arr[i].asSignedInt() );
}

void Test_operatorLessLess()
{
    EQ( 0, Str("").aSize() );
    EQ( 1, Str("x").aSize() );
    EQ( 2, Str("xx").aSize() );
    EQ( 3, Str("xxx").aSize() );

    EQ( 'a', Str("abc").aGet(0).asInt() );
    EQ( 'b', Str("abc").aGet(1).asInt() );
    EQ( 'c', Str("abc").aGet(2).asInt() );

    EQ( 0, ( NIL<<("") ).aSize() );
    EQ( 1, ( NIL<<("x") ).aSize() );
    EQ( 2, ( NIL<<("xx") ).aSize() );
    EQ( 3, ( NIL<<("xxx") ).aSize() );

    eq_vectors( Str("abc"), Str("a") << Str("b") << 'c' );
    eq_vectors( Str("abc"), Str("a") << Str("b") << "c" );
    eq_vectors( Str("abc"), Str("a") << 'b' << "c" );
    eq_vectors( Str("abc"), Str("a") << Str("b") << Str("c") );
    eq_vectors( Str("abc"), ( Str("a") << Str("b") ) << Str("c") );
    eq_vectors( Str("abc"), Str("a") << ( Str("b") << Str("c") ) );

    eq_vectors( Str("123"), NIL<<tri(123) );
    eq_vectors( Str("123"), NIL<< 123 );
    eq_vectors( Str("`xyz'"), NIL<<tri('x','y','z') );

}

void Test_tab_recall()
{
    tri x;
    x.setNewTab();
    tab *tp = x.asTab();

    for (int i = 0; i < 111; i++)
	tp->arr[i].setInt(i * i);
    for (int i = 0; i < 111; i++)
	EQ( tri(i*i), tp->recall( tri(i) ) );
    for (int i = 111; i < 222; i++)
	EQ( tri(), tp->recall( tri(i) ) );
}

void Test_tab_store()
{
    tri x;
    x.setNewTab();
    tab *tp = x.asTab();

    for (int i = 0; i < 111; i++)
	tp->store(tri().setInt(i), tri().setInt(i + i));

    const tab *ctp = x.asTab();
    for (int i = 0; i < 111; i++)
	EQ(i + i, ctp->arr[i].asInt());

    for (int i = 0; i < 111; i++)
	EQ( tri(i + i), tp->recall(tri(i)) );

    for (int i = 111; i < 222; i++)
	EQ( tri(), tp->recall(tri(i)) );
}

void Test_parseWords()
{
    tri a;
    a.setTCodeArrayFromString(" \t  \n ");
    EQ(0, a.aSize());

    tri b1;
    b1.setTCodeArrayFromString(" a ");
    EQ(1, b1.aSize());
    MUST(tri('a') == b1.aGet(0));
    EQ(tri::SYM | 0x61, b1.aGet(0).peek());

    tri b2;
    b2.setTCodeArrayFromString(" ab ");
    EQ(1, b2.aSize());
    MUST(tri('a', 'b') == b2.aGet(0));
    EQ(tri::SYM | 0x6162, b2.aGet(0).peek());

    tri b3;
    b3.setTCodeArrayFromString(" abc ");
    EQ(1, b3.aSize());
    MUST(tri('a', 'b', 'c') == b3.aGet(0));
    EQ(tri::SYM | 0x616263, b3.aGet(0).peek());

    tri b4;
    b4.setTCodeArrayFromString(" abcd ");
    EQ(1, b4.aSize());
    MUST(tri('a', 'b', 'c', 'd') == b4.aGet(0));
    EQ(tri::SYM | 0x61626364, b4.aGet(0).peek());


    tri c;
    c.setTCodeArrayFromString(" 555 ");
    EQ(1, c.aSize());
    EQ(555, c.aGet(0).asInt());
    EQ(tri().setInt(555).peek(), c.aGet(0).peek());
    MUST(tri().setInt(555) == c.aGet(0));

    tri d;
    d.setTCodeArrayFromString("777 wxyz");
    EQ(2, d.aSize());
    EQ(777, d.aGet(0).asInt());
    MUST(tri('w', 'x', 'y', 'z') == d.aGet(1));

    tri w;
    w.setTCodeArrayFromString("i WISH i WERE an Oscar Meyer Wiener!");
    EQ(8, w.aSize());
    MUST(tri('M', 'e', 'y', 'e') == w.aGet(6));

    w.setTCodeArrayFromString("one { two \"three\" four } five");
    EQ(3, w.aSize());

    EQ(tri('o', 'n', 'e') , w.aGet(0));

    EQ(3, w.aGet(1).aSize() );
    EQ( tri('t','w','o') , w.aGet(1).aGet(0) );
    MUST( w.aGet(1).aGet(1).isTab() );
    EQ( 5, w.aGet(1).aGet(1).aSize() );
    EQ( tri('f','o','u','r') , w.aGet(1).aGet(2) );

    EQ(tri('f', 'i', 'v', 'e') , w.aGet(2));

}

void Test_Terp()
{
    TriTerp terp;
    MUST( terp.home.isTab() );
    EQ( 0, terp.home.asTab()->arr.size() );

    EQ( terp.home.peek(), terp.f->lobby.peek() );

    terp.advanceFrame("Test_Terp");

    EQ( terp.home.peek(), terp.f->lobby.peek() );

    terp.retreatFrame("Test_Terp");

    EQ( terp.home.peek(), terp.f->lobby.peek() );

// first tcod test

    tri tcodes2(" tcod 23 42 + ");

    tri e2= terp.testEval( tcodes2, tri().setInt(111), tri().setInt(222) );
    EQ( NIL_GUTS, e2.peek() );

// tcod test with divide by 0

    tri tcodes3(" tcod 23 0 / ");

    tri e3= terp.testEval( tcodes3, tri().setInt(111), tri().setInt(222) );
    eq_vector_start( Str("division by zero"), e3 ); // TODO: match "divide by zero"
}

void Test_Terp_01() {

   //fprintf(stderr,"new %d del %d\n", count_new, count_delete );
    TriTerp terp;
}

void Test_Terp_02() {

   ////fprintf(stderr,"new %d del %d\n", count_new, count_delete );
    TriTerp terp;
    tri code( " tcod anew " );
}

void Test_Terp_03() {

   //fprintf(stderr,"new %d del %d\n", count_new, count_delete );
    TriTerp terp;
    tri code( " tcod " );
    tri z= terp.testEval( code, NIL, NIL );
}

void Test_Terp_anew() {

    {
    TriTerp terp;
	tri code( " tcod anew " );
	tri z= terp.testEval( code, tri(111), tri(222) );
	EQ( tri(), z );
	MUST( terp.pop().isTab() );
	EQ( tri(222), terp.pop() );
	EQ( tri(111), terp.pop() );
    }
}
void Test_Terp_10() {
   //fprintf(stderr,"new %d del %d\n", count_new, count_delete );
    {
    TriTerp terp;
	tri code( " tcod 13 aget " );
	tri a; a.setNewTab(); a.aPut( 13, 456 );
	tri z= terp.testEval( code, a, a);
	EQ( tri(), z );
	EQ( tri(456), terp.pop() );
	EQ( a, terp.pop() );
    }
   //fprintf(stderr,"new %d del %d\n", count_new, count_delete );
}
void Test_Terp_11() {

    {
    TriTerp terp;

	tri a; a.setNewTab();

	tri z= terp.testEval( tri("tcod 11 aput "), tri(888), a);
	EQ( NIL, z );

	z= terp.testEval( tri("tcod 11 aget ") , NIL, a);
	EQ( NIL, z );

	EQ( tri(888), terp.pop() );
    }

}
void Test_Terp_12() {

    {
    TriTerp terp;
	tri code( " tcod 101 sto/ 0 202 sto/ 1 " );
	tri z= terp.testEval( code, tri('a','a','a'), tri('b','b','b') );
	EQ( tri(), z );
	EQ( tri(202), terp.pop() );
	EQ( tri(101), terp.pop() );
    }
}
void Test_Terp_13() {

    {
    TriTerp terp;
	tri code( " tcod 101 sto/ 0 202 sto/ 1 home  rcl/ 0 rcl/ 1   home " );
	tri z= terp.testEval( code, tri('a','a','a'), tri('b','b','b') );
	EQ( tri(), z );
	EQ( terp.home, terp.pop() );
	EQ( tri(202), terp.pop() );
	EQ( tri(101), terp.pop() );
	EQ( terp.home, terp.pop() );
	EQ( tri(202), terp.pop() );
	EQ( tri(101), terp.pop() );
	//EQ( SYM4('h','o','m','e'), terp.home.asTab()->meta->tag );
    }
}
void Test_Terp_14() {

    {
    TriTerp terp;
	tri dbl('d','b','l');
	tri doubleCode( " tcod  rcl/ 2  rcl/ 2  + ret " );
	terp.home.asTab()->store( dbl, doubleCode );

	tri code( " tcod   home imm/ dbl   imm/ 1024   cal/ 3 " );
	tri z= terp.testEval( code, tri('a','a','a'), tri('b','b','b') );
	EQ( tri(), z );
	EQ( tri(2048), terp.pop() );
    }
}
void Test_Terp_15() {

    {
    TriTerp terp;
	tri dbl('d','b','l');
	tri doubleCode( " tcod  rcl/ 2  rcl/ 2  + ret " );
	terp.home.asTab()->store( dbl, doubleCode );

	tri code( " tcod   home imm/ dbl   home imm/ dbl     imm/ 1024  cal/ 3 cal/ 3 " );
	tri z= terp.testEval( code, tri('a','a','a'), tri('b','b','b') );
	EQ( tri(), z );
    }
}
void Test_Terp_16() {

    {
    TriTerp terp;
	tri code( " tcod if { tcod imm/ yes } { tcod imm/ no } " );
	tri z= terp.testEval( code, tri('a','a','a'), tri( 1 ) );
	EQ( tri(), z );
	EQ( tri('y','e','s'), terp.pop() );
	z= terp.testEval( code, tri('a','a','a'), tri( 0 ) );
	EQ( tri(), z );
	EQ( tri('n','o'), terp.pop() );
    }
}
void Test_Terp_17() {

    {
    TriTerp terp;
	// factorial using recursion
	tri fact('f','a','c','t');
	tri factCode( " tcod  rcl/ 2 if { tcod home imm/ fact  rcl/ 2 1 - cal/ 3  rcl/ 2 * } { tcod 1 } ret ");
	terp.home.asTab()->store( fact, factCode );

	tri code( " tcod   home imm/ fact   5   cal/ 3 " );
	tri z= terp.testEval( code, tri('a','a','a'), tri('b','b','b') );
	EQ( tri(), z );
	EQ( tri(120), terp.pop() );
    }
}
void Test_Terp_18() {

    {
    TriTerp terp;
	// triangle using recursion
	tri tria('t','r','i','a');
	tri triaCode( " tcod  rcl/ 2 if { tcod home imm/ tria  rcl/ 2 1 - cal/ 3  rcl/ 2 + } { tcod 0 } ret ");
	terp.home.asTab()->store( tria, triaCode );

	tri code( " tcod   home imm/ tria   5   cal/ 3 " );
	tri z= terp.testEval( code, tri('a','a','a'), tri('b','b','b') );
	EQ( tri(), z );
	EQ( tri(15), terp.pop() );
    }
}
void Test_Terp_19() {

    {
    TriTerp terp;
	// triangle using while and return
	tri tria('t','r','i','a');
	tri triaCode( " tcod 0 0 while { tcod  rcl/ 3 rcl/ 2 > if { tcod rcl/ 4 ret } { tcod }   rcl/ 3 rcl/ 4 + sto/ 4 rcl/ 3  1 +  sto/ 3  55555 }  " );
	terp.home.asTab()->store( tria, triaCode );

	tri code( " tcod   home imm/ tria   5   cal/ 3 " );
	tri z= terp.testEval( code, tri('a','a','a'), tri('b','b','b') );
	EQ( tri(), z );
	EQ( tri(15), terp.pop() );
    }
}
void Test_Terp_20() {

    {
    TriTerp terp;
	// triangle using while and break
	tri tria('t','r','i','a');
	tri triaCode( " tcod 0 0 while { tcod   rcl/ 3 rcl/ 2 > brk    rcl/ 3 rcl/ 4 + sto/ 4    rcl/ 3  1 +  sto/ 3  55555 }  rcl/ 4 " );
	terp.home.asTab()->store( tria, triaCode );

	tri code( " tcod   home imm/ tria   22   cal/ 3 " );
	tri z= terp.testEval( code, tri('a','a','a'), tri('b','b','b') );
	EQ( tri(), z );
	EQ( tri( 22*23/2 ), terp.pop() );
    }

}
   
void Test_Compiler_1() {
    tri a= CompileString( tri(" 3 ") );

    //fprintf(stderr, "\n");
    //a.to(stderr); 
    //fprintf(stderr, "<-a\n");
    if (verbose) dumpVec(a, "a" );
    EQ( 2, a.aSize() );
    EQ( tri(3), a.aGet(1) );

    tri b= CompileString( tri(" 1 + 2 - 3 * 4 / 5 ") );
    //fprintf(stderr, "\n");
    //b.to(stderr); 
    //fprintf(stderr, "<-b\n");
    if (verbose) dumpVec(b, "b" );

    tri expected(" tcod   1 2 + 3 - 4 * 5 /   ");
    eq_vectors( expected, b );

}
   
void Test_Compiler_slot() {
//++verbose;
    tri code= CompileString( tri(" a + 7 ") );
    if(verbose) dumpVec(code, "code: a + 7" );
    eq_vectors( tri("tcod home imm/ a aget 7 + ") , code );

    code= CompileString( tri(" a . b + 7 ") );
    if(verbose) dumpVec(code, "code: a . b + 7" );
    eq_vectors( tri("tcod home imm/ a aget imm/ b aget 7 + ") , code );

    code= CompileString( tri(" Home . a = 66 ") );
    if(verbose) dumpVec(code, "code: Home . a = 66 " );
    eq_vectors( tri("tcod 66 home imm/ a aput   ") , code );

    code= CompileString( tri(" lcl = 66 ") , "lcl" );
    if(verbose) dumpVec(code, "code: lcl = 66" );
    eq_vectors( tri("tcod  66 sto/ 0    ") , code );

    code= CompileString( tri(" a . b = 66 ") );
    if(verbose) dumpVec(code, "code.." );
    eq_vectors( tri("tcod 66 home imm/ a aget imm/ b aput    ") , code );

    code= CompileString( tri(" lcl . b = 66 "), "lcl");
    if(verbose) dumpVec(code, "code.." );
    eq_vectors( tri("tcod 66 rcl/ 0 imm/ b aput    ") , code );
//--verbose;
}

void Test_Compiler_locals() {

    tri src(" 1 + b - c * d / 5 ");

    tri dest("");
    Compiler c( src, dest );
    c.localVars.setNewTab();

    EQ( (word)-1, (word) c.findLocal( tri('a') ) );
    EQ( (word)-1,   (word) c.findLocal( tri('b') ) );
    EQ( (word)-1,   (word) c.findLocal( tri('c') ) );
    EQ( (word)-1,   (word) c.findLocal( tri('y') ) );

    c.addLocal( tri('a') );

    EQ( 0,   (word) c.findLocal( tri('a') ) );
    EQ( (word)-1,   (word) c.findLocal( tri('b') ) );
    EQ( (word)-1,   (word) c.findLocal( tri('c') ) );
    EQ( (word)-1,   (word) c.findLocal( tri('y') ) );

    c.addLocal( tri('a') );
    c.addLocal( tri('a') );
    c.addLocal( tri('b') );
    c.addLocal( tri('b') );
    c.addLocal( tri('c') );
    c.addLocal( tri('c') );

    EQ( 0,   (word) c.findLocal( tri('a') ) );
    EQ( 1,   (word) c.findLocal( tri('b') ) );
    EQ( 2,   (word) c.findLocal( tri('c') ) );
    EQ( (word)-1,   (word) c.findLocal( tri('y') ) );

    c.addLocal( tri('d') );
    c.addLocal( tri('d') );

    Expr* ex= c.parseExpr();
    //ex->print(stderr);
    ex->emitGet();
    //dumpVec(dest);

    tri expected("  1 rcl/ 1 + rcl/ 2 - rcl/ 3 * 5 /  ");
    eq_vectors( expected, dest );

    delete ex;
}

void XXXXXXXXXXXXXXTest_Compiler_bad() {
    
    tri src(" 1 + ");
    tri dest("");
    Compiler c( src, dest );

    Expr* ex= NULL;
    ISBAD( ex= c.parseExpr() )
    EQ( 0 , (word)ex );

}

void Test_Compiler_bad_empty() {
#ifdef DEBUG
    
    tri src("    ");
    tri dest("");
    Compiler c( src, dest );

    Expr* ex= NULL;
    ISBAD( ex= c.parseExpr() )
    MUST( ex->isError() );
    delete ex;
#endif
}

void Test_Compiler_fun_02_invoke() {

    tri program= tri::NewCodVec(" fun ( x ) { x + x } ");
    tri code(" tcod ");

    Compiler piler( program, code );

    Expr* z= piler.parseExpr();
    z->emitGet();
    IFD( if (verbose) z->print(); )
    delete z;
    
    //if (verbose) dumpVec(code, "2: code" );		// tcod imm/ <fun>
    //if (verbose) dumpVec(code.aGet(2), "2: code[2]" );

    //++verbose; 
    TriTerp terp;
    tri e2= terp.testEval( code, tri().setInt(111), tri().setInt(222) );
    EQ( NIL, e2 );

    tri fun= terp.pop();
    EQ( 8, fun.aSize() );
    EQ( tri('t','c','o','d'), fun.aGet(0) );
    EQ( tri('l','c','l','/'), fun.aGet(1) );
    EQ( tri('+'), fun.aGet(7) );
    //--verbose; 
}

void Test_Compiler_fun_03_assign() {

    tri program= tri::NewCodVec("Home . mul2 =  fun ( x ) { x + x } ; mul2 ( 10 ) ");
    tri code(" tcod ");

    Compiler piler( program, code );

    Error* error= NULL;
    Expr* z= piler.parseExprList( &error );
    CANT( error );
    z->emitGet();
    IFD( if (verbose) z->print(); )
    delete z;
    
    if (verbose) dumpVec(code, "3: code" );		// tcod imm/ <fun>
    if (verbose) dumpVec(code.aGet(5), "3: fun()" );

    //++verbose; 
    TriTerp terp;
    tri e2= terp.testEval( code, tri().setInt(111), tri().setInt(222) );
    EQ( NIL, e2 );

    EQ( tri(20), terp.pop() );
    //--verbose; 
}

void Test_Compiler_if_0() {

    tri program= tri::NewCodVec(" if ( 0 ) { 1001 } else { 1002 } " );
    tri code(" tcod ");

    Compiler piler( program, code );

    Error* error= NULL;
    Expr* z= piler.parseExprList( &error );
    CANT( error );
    z->emitGet();
    IFD( if (verbose) z->print(); )
    delete z;
    
    if (verbose) dumpVec(code, "if_1: code" );		// tcod imm/ <fun>

    TriTerp terp;
    tri e2= terp.testEval( code, tri().setInt(111), tri().setInt(222) );
    EQ( NIL, e2 );

    EQ( tri(1002), terp.pop() );
}

void Test_Compiler_if_1() {

    tri program= tri::NewCodVec(" if ( 1 ) { 1001 } else { 1002 } " );
    tri code(" tcod ");

    Compiler piler( program, code );

    Error* error= NULL;
    Expr* z= piler.parseExprList( &error );
    CANT( error );
    z->emitGet();
    IFD( if (verbose) z->print(); )
    delete z;
    
    if (verbose) dumpVec(code, "if_1: code" );		// tcod imm/ <fun>

    TriTerp terp;
    tri e2= terp.testEval( code, tri().setInt(111), tri().setInt(222) );
    EQ( NIL, e2 );

    EQ( tri(1001), terp.pop() );
}

void Test_Compiler_loop_1() {

    //++verbose; 
    tri program= tri::NewCodVec(" nil nil nil i = 10 ; z = 0; ; loop { if ( i ) { z = ( z + i ) } else { brk } ; i = ( i - 1 )  ; } z " );
    tri code(" tcod ");

    Compiler piler( program, code );

    Error* error= NULL;
    Expr* z= piler.parseExprList( &error );
    CANT( error );
    z->emitGet();
    IFD( if (verbose) z->print(); )
    delete z;
    
    if (verbose) dumpVec(code, "loop_1: code" );		// tcod imm/ <fun>

    TriTerp terp;
    tri e2= terp.testEval( code, tri().setInt(111), tri().setInt(222) );
    EQ( NIL, e2 );

    EQ( tri( 55 ), terp.pop() );
}

tri EvalTripe(char* program) {
    tri tripe= tri::NewCodVec(program);
    IFD( if (verbose) say << "\n==Codes: " << tripe; )
    TriTerp terp;
    tri e= terp.testEval( tripe, NIL, NIL );
    IFD( if (verbose) say << "\n==Status: " << e; )
    EQ( NIL, e );
    tri z= terp.pop();
    IFD( if (verbose) say << "\n==Result: " << z; )
    return z;
}

tri Eval(char* program) {
    IFD( if (verbose) say << "\n==Program: " << program; )
    tri code("tcod");
    Compiler piler( tri::NewCodVec(program), code );
    Error* error= NULL;
    Expr* expr= piler.parseExprList( &error );
    CANT( error );
    expr->emitGet();
    IFD( if (verbose) say << "\n==Expr: "; )
    IFD( if (verbose) expr->print(); )
    IFD( if (verbose) say << "\n==Codes: " << code; )
    TriTerp terp;
    tri e= terp.testEval( code, NIL, NIL );
    IFD( if (verbose) say << "\n==Status: " << e; )
    EQ( NIL, e );
    tri z= terp.pop();
    IFD( if (verbose) say << "\n==Result: " << z; )
    delete expr;
    return z;
}

void Test_Fact() {
    EQ( tri(120) , Eval("Fact = fun ( n ) { if ( n ) { n * ( Fact ( n - 1 ) ) } else { 1 } } ; Fact ( 5 )") );
}

void Test_Fib() {
    EQ( tri(13) , Eval(
	"Fib = fun ( x ) { if   ( x < 2 )   { x }   else   { ( Fib ( x - 2 ) ) + ( Fib ( x - 1 ) ) } } ; Fib ( 7 ) "
    ) );
}

void Test_anew() {
    tri x = Eval( "anew ( )");
    MUST( x.isTab() );
    EQ( 0, x.aSize() );
}

void Test_aget_tripe() {
    EQ( tri(25) , EvalTripe("tcod  anew sto/ 1  25 rcl/ 1 5 aput  9 rcl/ 1 3 aput   rcl/ 1 5 aget" ) );
}

void Test_aget() {
    EQ( tri(25) , Eval("a = anew ( ) ; aput ( a , 5 , 25 ) ; aput ( a , 3 , 9 ) ; aget ( a , 5 )" ) );
}

void Test_putup_getup1() {
    // this one explicitly clears the 'a' slot, to break loop
    EQ( tri(1) , Eval(" a = anew ( ) ; putup ( a , Home ) ; z = ( Home == getup ( a ) ) ; a = nil ; z " ) );
}
void Test_putup_getup2() {
    // this one requires terp to reset home, to break loop
    EQ( tri(1) , Eval(" a = anew ( ) ; putup ( a , Home ) ;  ( Home == getup ( a ) )  " ) );
}

void Test_ZZZ_stack_assign() {
//++verbose;
    TriTerp terp;
    //say << "sp=" << terp.stackSize();

    tri code= CompileString(" 1111 ",  "a" );
    err e= terp.testEval( code, NIL, NIL );
    EQ( 4, terp.stackSize() );

    code= CompileString(" a = 64 ",  "a" );
    e= terp.testEval( code, NIL, NIL );
    EQ( 4, terp.stackSize() );
//--verbose;
}

void Test_atoi()
{
    EQ( 54321, atoi("54321") );
    EQ( (u4)-1234, (u4)atoi("-1234") );
}

void Test_itoa()
{
    char buf[33];
    itoa( 11235, buf );
    EQ( '1', buf[0] );
    EQ( '1', buf[1] );
    EQ( '2', buf[2] );
    EQ( '3', buf[3] );
    EQ( '5', buf[4] );
    EQ( '\0', buf[5] );

    itoa( -789, buf );
    EQ( '-', buf[0] );
    EQ( '7', buf[1] );
    EQ( '8', buf[2] );
    EQ( '9', buf[3] );
    EQ( '\0', buf[4] );
}

void DisasmAssignNil(tri* p)
{
    *p = tri();
}

void DisasmAssignBlah(tri* p)
{
    *p = tri('b','l','a','h');
}

tri DisasmReturn888(tri* p)
{
    return tri(888);
}

#endif

// END
