
//#include <stdio.h>
//#include <unistd.h>

#ifndef NULL
#define NULL 0
#endif

#ifndef NOREFC
#define REFC 1
#endif

///////////// runtime

extern int PutcStdout( int c );
extern int PutcStderr( int c );

extern int PutsStdout( const char[], int len );
extern int PutsStderr( const char[], int len );


void TriInit(void);
void Say(int level, const char *s);  // debugging prints
void Bomb(const char* problem);   // aborts
void Bad(void);                   // breakpoint here
void MemCheck(void); 

extern int SYSopen( const char* name, int mode );
extern int SYSread(int fd, char* buf, int sz);
extern int SYSwrite( int, const void*, int );
extern int SYSclose(int fd);

#ifndef MINIMAL

#define SYSopen open
#define SYSread read
#define SYSwrite write
#define SYSclose close

#endif

////////////////////  integers

typedef unsigned char u1;
typedef unsigned short u2;
typedef unsigned int u4;
typedef   signed char s1;
typedef   signed short s2;
typedef   signed int s4;

typedef u4 word;  // must be unsigned

////////////////////  asserts

extern "C" {

#include <assert.h>

    void sync(void);  // for breakpoints
    void abort(void);
    char* getenv( const char* );
    int atoi( const char* );
    char* itoa( int i, char* buf ); // returns ptr past end
    int strlen( const char* );
}

extern int expect_error;
extern int count_error;
extern int count_expected_error;


extern int count_new;
extern int count_delete;
extern int count_new_array;
extern int count_delete_array;

extern int verbose;


#ifdef DEBUG

#define DBG if(verbose)fprintf

#define MUST(a) assert_must(a,__STRING(a),__LINE__,__FILE__,__PRETTY_FUNCTION__)
#define CANT(a) assert_must((!(a)),__STRING((!(a))),__LINE__,__FILE__,__PRETTY_FUNCTION__)
#define EQ(a,b) assert_eq(a,b,__STRING(a),__STRING(b),__LINE__,__FILE__,__PRETTY_FUNCTION__)
#define NE(a,b) assert_ne(a,b,__STRING(a),__STRING(b),__LINE__,__FILE__,__PRETTY_FUNCTION__)
#define ISBAD(a) { int old= count_expected_error; expect_error=1; (a); expect_error=0; MUST(old<count_expected_error); }

const int TOOBIG = 100000;  // nothing should get bigger

#else

#define DBG if(0)fprintf

//#define MUST(a)  ( (!(a)) ? Bomb("Assertion failed: MUST" ),0 : 0 )
//#define CANT(a)  ( ( (a)) ? Bomb("Assertion failed: CANT" ),0 : 0 )
//#define EQ(a,b)  ( ((a)!=(b)) ? Bomb("Assertion failed: EQ" ),0 : 0 )
//#define NE(a,b)  ( ((a)==(b)) ? Bomb("Assertion failed: NE" ),0 : 0 )
#define ISBAD(a) /*nothing*/

inline void MUST(bool a) { if (!a) Bomb("ASSERTION FAILED"); }
inline void CANT(bool a) { if ( a) Bomb("ASSERTION FAILED"); }

inline void EQ(int a, int b) { if ( a != b ) Bomb("ASSERTION FAILED"); }
inline void NE(int a, int b) { if ( a == b ) Bomb("ASSERTION FAILED"); }

const int TOOBIG = 0x0fffFFFF;  // nothing should get bigger

#endif

class tri;
class tab;
class vec;
class TriTerp;
struct Meta;

typedef tri err;  // when NIL means good, and else means error

bool assert_must(bool a, const char* A, int line, const char* file, const char* func);
bool assert_eq(word a, word b, const char* A, const char* B, int line, const char* file, const char* func);
bool assert_ne(word a, word b, const char* A, const char* B, int line, const char* file, const char* func);
bool assert_eq(tri a, tri b, const char* A, const char* B, int line, const char* file, const char* func);
bool assert_ne(tri a, tri b, const char* A, const char* B, int line, const char* file, const char* func);

void CheckDeep(tri x, int depth= 8);

#ifdef DEBUG
struct Sayer {
	Sayer& operator << (int a);
	Sayer& operator << (char a);
	Sayer& operator << (const char* a);
	Sayer& operator << (const tri& a);
	Sayer& operator << (const tab& a);
	Sayer& operator << (const vec& a);
	Sayer& operator << (const TriTerp& a);
	Sayer& operator << (class Expr* a);

	int level;
};
#else
struct Sayer {
	Sayer& operator << (int a) { return *this; }
	Sayer& operator << (char a) { return *this; }
	Sayer& operator << (const char* a) { return *this; }
	Sayer& operator << (const tri& a) { return *this; }
	Sayer& operator << (const tab& a) { return *this; }
	Sayer& operator << (const vec& a) { return *this; }
	Sayer& operator << (const TriTerp& a) { return *this; }
	Sayer& operator << (class Expr* a) { return *this; }
};
#endif

extern Sayer say;
extern Sayer c0;
extern Sayer c1;
extern Sayer c2;
extern Sayer c3;
extern Sayer c4;
extern Sayer c5;
extern bool runUnitTests;

///////////////////  classes

#define SYM1(a) ( tri::SYM | (u1)(a) )
#define SYM2(a,b) ( tri::SYM | ((u1)(a)<<8) | (u1)(b) )
#define SYM3(a,b,c) ( tri::SYM | ((u1)(a)<<16) | ((u1)(b)<<8) | (u1)(c) )
#define SYM4(a,b,c,d) ( tri::SYM | ((u1)(a)<<24) | ((u1)(b)<<16) | ((u1)(c)<<8) | (u1)(d) )

#define	NIL_GUTS	SYM3('n','i','l') 
#define	NIL		tri('n','i','l')


#ifdef DEBUG
#  define MAGIC  0xae9b5366
#  define	IFD(X) X
#else
#  define	IFD(X) /*nothing*/
#endif

/////////////////////////////////////////////////////////

class vec {
private:
    int	    siz;
    int	    cap;
    tri*    d;

    static word const_zero;

    void do_grow(int goal);
    void grow(int goal) {
	// if we already are at goal, ok
	if (cap<goal) do_grow(goal);
    }
    void operator= (const vec& FORBIDDEN);

public:
    ~vec();
    vec() : siz(0), cap(0) , d(0) {}
    tri& operator [] (int i) const ;
    tri& operator [] (int i) ;
    int capacity() const { return cap; }
    int size() const { return siz; }
    void setSize(int);
};

class tab {

friend class tri;

private:
#ifdef REFC
    int	    refs;
#endif

    tab(tri* setMe);
    void operator= (const tab& FORBIDDEN);

public:

#ifdef MAGIC
    word    magic;
#endif

    Meta*   meta;
    word    upWord;
    tri*    up() { return (tri*) &upWord; }

    vec	    arr;
    vec	    dict;

    static long    countCtors;
    static long    countDtors;

    ~tab();
#ifdef MAGIC
    void checkMagic() const;
#else
    void checkMagic() const {}
#endif

    void reset();

    int	aSize() const { checkMagic(); return arr.size(); }
    int	dSize() const { checkMagic(); return dict.size(); }

    tri aGet(int i) const;
    void aPut(int i, tri x);

    tri recall(const tri& tindex) const;
    tri recall(word windex) const;

    void store(const tri& tindex, const tri& x) ;
    void store(word windex, const tri& x) ;

    void aAdd(tri item);

    void link( void* by )
    {
#ifdef DEBUG
	checkMagic();
#ifdef REFC
	MUST( refs>=0 );
	MUST( refs<=TOOBIG );
#endif
#endif
#ifdef REFC
	++refs;
#endif
    }

    void unlink( void* by )
    {
#ifdef DEBUG
	checkMagic();
#ifdef REFC
	MUST( refs>0 );
	MUST( refs<=TOOBIG );
#endif
#endif

#ifdef REFC
	--refs;
	if (refs == 0) {
	    delete this;
	}
#endif
    }
 
};




class tri {  // the tri-modal trithorpe thingy
    

private:
	word  guts;

	static const int NUM_TAG_BITS = 2;

	static const word BASIS = 1 << ( 8 * sizeof(word) - NUM_TAG_BITS );
	static const word LO = BASIS-1;
	static const word HI = ~LO;

public:
	tri() : guts( NIL_GUTS ) { link(); }
	tri(const tri& copyMe) : guts(copyMe.guts) { link(); }
	tri(tri& copyMe) : guts(copyMe.guts) { link(); }

	tri(int x) : guts(LO&x | INT&HI) { }
	tri(word x) : guts(LO&x | INT&HI) { }

	tri(char a) : guts( SYM1(a) ) { IFD( MUST(isSym()) ); }
	tri(char a, char b) : guts( SYM2(a,b) ) { IFD( MUST(isSym()) ); }
	tri(char a, char b, char c) : guts( SYM3(a,b,c) ) { IFD( MUST(isSym()) ); }
	tri(char a, char b, char c, char d) : guts( SYM4(a,b,c,d) ) { IFD( MUST(isSym()) ); }

	tri(const char* tcodeScript);

	static tri NewCodVec(const char*);
	static tri NewDeepCodVec(const char*);

	static const word INT =  BASIS * 0;
	static const word SYM =  BASIS * 1;
	static const word UNK =  BASIS * 2;
	static const word TAB =  BASIS * 3;


	static word isWhat(word w) { return (w&HI); }
	static bool isInt(word w) { return (w&HI) == INT; }
	static bool isSym(word w) { return (w&HI) == SYM; }
	static bool isUnk(word w) { return (w&HI) == UNK; }
	static bool isTab(word w) { return (w&HI) == TAB; }

	word isWhat() const { return (guts&HI); }
	bool isInt() const { return (guts&HI) == INT; }
	bool isSym() const { return (guts&HI) == SYM; }
	bool isUnk() const { return (guts&HI) == UNK; }
	bool isTab() const { return (guts&HI) == TAB; }

	word peek() const { return guts; }

	bool operator == (const tri& that) const { return this->guts == that.guts; }
	bool operator != (const tri& that) const { return this->guts != that.guts; }

	tri& operator = (const tri& copyMe) { unlink(); this->guts= copyMe.guts; link(); return *this; }

	int	aSize() const { return asTab()->aSize(); }
	int	dSize() const { return asTab()->dSize(); }

	tri   aGet(int i) const;
	void  aPut(int i, tri x);
	void  aAdd(tri item);

	void  aStore(tri key, tri value) { asTab()->store(key,value); }
	tri   aRecall(tri key) { return asTab()->recall(key); }


	static word asInt(word w) {

	    MUST( isInt(w) );

	    return w & LO; 
	}

	static tab* asTab(word w) {

	    MUST( isTab(w) );

	    tab* z= (tab*)((w & LO)<<2);
	    z->checkMagic();
	    return z;
	}

	word asComparisonWord() const {

	    // this is tricky.  We bias the word z so that positive Signed Int are the largest 
	    // values returned, with negative Signed int beneath them, and all non-ints less than all ints.
	    // Subtracting ( BASIS >> 1 ) does the trick.

	    return peek() - ( BASIS >> 1 );
	}

	int asSignedInt() const {

	    MUST( isInt() );

	    int z= (int) ( guts & LO ); 
	    return ( z<<NUM_TAG_BITS ) >> NUM_TAG_BITS;
	}

	word asInt() const {

	    MUST( isInt() );

	    return guts & LO; 
	}

	tab* asTab() const {

	    MUST( isTab() );

	    tab* z= (tab*)((guts & LO)<<2);
	    z->checkMagic();
	    return z;
	}

	tri& setInt(word x) {
	    unlink();
	    guts= ( x&LO | INT&HI ); 
	    return *this;
	}
	tri& setTab(tab* x) {   // called back from tab::tab
	    x->checkMagic();
	    unlink();

	    MUST( x );

	    guts= ( (u4(x)>>2)&LO | TAB&HI ); 
	    linkUnchecked();

	    return *this;
	}
	tri& setNewTab() {
	    /*tab* unused=*/ new tab( this );   // will call back, to setTab()
	    return *this;
	}

	tri& setSym(char a) { setSym('@',0,0,a); return *this; }
	tri& setSym(char a, char b) { setSym('@',0,a,b); return *this; }
	tri& setSym(char a, char b, char c) { setSym('@',a,b,c); return *this; }
	tri& setSym(char a, char b, char c, char d);
	tri& setSym(word w);

	tri& setTag(tri tag);

	~tri() { unlink(); }
	void linkUnchecked() { 
	    //DBG(stderr, "tri:link %x %x %x\n", this, guts, guts*4); 
	    asTab()->link(this);
	}
	void link() { 
#ifdef REFC
	    //DBG(stderr, "tri:link %x %x %x\n", this, guts, guts*4); 
	    if (isTab()) { asTab()->link(this); }
#endif
	}
	void unlink() { 
#ifdef REFC
	    //DBG(stderr, "tri:unlink %x %x %x\n", this, guts, guts*4); 
	    if (isTab()) { asTab()->unlink(this); }
#endif
	    guts= 0;
	}

    tri recall(const tri& tindex) const ;


    tri& setTCodeItemFromString(const char* s, int begin, int end);
    tri& setTCodeArrayFromString(const char* s, int *returnLastUsed= NULL, bool deeply= true);


    void toString( char buf[] ) const;

	tri& operator << (int a);
	tri& operator << (word a);
	tri& operator << (char a);
	tri& operator << (unsigned char a);
	tri& operator << (const char* a);
	tri& operator << (const tri& a);
	tri& operator << (const tab& a);
	tri& operator << (const vec& a);
	tri& operator << (const TriTerp& a);
	tri& operator << (class Expr* a);
};


inline    tri& vec::operator [] (int i) const {

	if (i<cap) {
	    void* z= (char*)d + i * sizeof(word);
	    return * (tri*) z;
	} else {
		return *(tri*)(&const_zero);
	}
}

inline    tri& vec::operator [] (int i) {
	grow(i+1);
	if (siz<i+1) siz= i+1;

	void* z= (char*)d + i * sizeof(word);
	return * (tri*) z;
}

inline tri tab::aGet(int i) const { checkMagic(); return arr[i]; }
inline void tab::aPut(int i, tri x) { checkMagic(); arr[i]= x; }
inline void tab::aAdd(tri item) { checkMagic(); aPut( aSize(), item ); }

inline tri tri::aGet(int i) const { return asTab()->aGet(i); }
inline void tri::aPut(int i, tri x) { asTab()->aPut(i,x); }
inline void tri::aAdd(tri item) { asTab()->aAdd( item ); }

#ifndef DEBUG
inline void EQ(tri a, tri b) { if ( a != b ) Bomb("ASSERTION FAILED"); }
inline void NE(tri a, tri b) { if ( a == b ) Bomb("ASSERTION FAILED"); }
#endif

/////////////////////////////////////////////////////////
//
//   Strings.........

//extern err PutsStdout( tri );
//extern err PutsStderr( tri );

bool IsPrintableAsciiObj(tri x);
char* StrDupIfPrintableAsciiObj(tri x);

int StrLen(const char* s);
int StrSum(const char* s);
char* StrDup (const char* s);
int StrCmp(const char* a, const char* b);

inline bool StrEq(const char* a, const char* b) { return !StrCmp(a,b);}

tri	Str(const char* s);        // create new String tri
bool 	IsPritableAsciiObj(tri x);
tri	Intern(const char* s);

struct InternStr {
	tri		obj;
	const char*	str;
	InternStr*	next;

	InternStr(tri _obj, const char* _str, InternStr* _next);
};

class InternTable {

#define INTERN_MASK 255
	InternStr*	bucket[INTERN_MASK+1];
public:
	InternStr*	intern(const char* s);
};

extern InternTable It;

/////////////////////////////////////////////////////////

inline bool IsTrue(tri x) { word w= x.peek(); return w!=0 && w!=SYM3('n','i','l'); }
inline bool IsFalse(tri x) { word w= x.peek(); return w==0 || w==SYM3('n','i','l'); }

// met = meta table for tab
struct Meta {
    tri  tag;
    void (*finalizer)(tab*);
};


#define MAX_FRAMES 99
#define MAX_STACK  (MAX_FRAMES * 10)

struct TriFrame {  // a stack frame 
    static const int STOP= 1;
    static const int CATCH= 2;
    static const int FUNC= 4;
    static const int LOOP= 8;

    int	    fflags;
    tri	    kp;
    int    pc;
    tri*    bp;
    tri*    savedSp;
    tri	    lobby;

    TriFrame();
};


class TriTerp {  // main inTerpreter structure

    struct TriFrame frames[MAX_FRAMES];

    tri*    sp;
    tri	    stack[MAX_STACK];

public:

    struct TriFrame *f;
    tri	    home;

// constants
    const tri TCODE_MAGIC;

// cotr
    TriTerp();
    ~TriTerp();
    void init();    // in case of no constructor

// frames
    void advanceFrame(const char* why);
    void retreatFrame(const char* why);

    err	testEval( tri script, tri arg1, tri arg2 );

    int stackSize() { return sp - stack; }

private:
    tri doLoop();
    void doInt( const tri& );
    err doSym( tri );
    err doTab( tri );
    err doUnk( word );
    err doCall();
    err doReturn();
    err doStop();
    err doIf();
    err doWhile();
    err doBreak();
    err doPutc();
    err doPuts();
    err doLclInfo();
    err doAbort();
    err doGetUp();
    err doPutUp();
    
    tri pcImmediate();  // returns immediate entry at program counter & increments pc

public:
    void dumpStack() const;

    // stack operations:
    void TriTerp::push(const tri& x) {
#ifdef DEBUG
	if (verbose) { say<< "PUSH <--" << x << "\n"; }
#endif
	*sp= x;
	++sp;
#ifndef DANGER
	MUST(sp<stack+MAX_STACK);
#endif
    }
    tri TriTerp::pop() {
#ifndef DANGER
	MUST(sp>stack);
#endif
	--sp;
#ifdef DEBUG
	if (verbose) { say<< "POP -->" << *sp << "\n"; }
#endif
	return *sp;
    }
     
    tri* TriTerp::peek() {
#ifndef DANGER
	MUST(sp>stack);
#endif
#ifdef DEBUG
	if (verbose) { say<< "PEEK -->" << sp[-1] << "\n"; }
#endif
	return sp-1;
    }
     
};


tri CompileString(tri x, char* locals= "" );
void dumpVec(tri x, const char* s, int depth=4);

